#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <stddef.h>
#include <stdarg.h>

static char const prog[]="svclean-shutdown";

static int status=0;
static void setstatus(void) {
  int const es=errstat(errno);
  if (status==0 || es==100) status=es;
}

static void printerr(char const* s0, char const* s1, char const* s2) {
  int error=errno;
  buffer_puts(buffer_2, prog);
  buffer_puts(buffer_2, err_);
  buffer_puts(buffer_2, err_unable);
  buffer_puts(buffer_2, s0);
  if (s1!=null) {
    buffer_puts(buffer_2, s1);
    if (s2!=null)
      buffer_puts(buffer_2, s2);
  }
  buffer_puts(buffer_2, err_);
  buffer_puts(buffer_2, error_str(error));
  buffer_puts(buffer_2, "\n");
  buffer_flush(buffer_2);
}

static void printerr_die(char const* s0, char const* s1, char const* s2) {
  setstatus();
  printerr(s0, s1, s2);
  _exit(status);
}

typedef struct srvlist srvlist;
struct srvlist {
  srvlist* next;
  unsigned int up:1, err:1, islog:1;
  char name[1];
};
#define srvlist_size(len) (offsetof(srvlist, name)+(len))

static srvlist* services;
static int servicesdir;
static char const* servicesdirname;

static void create_shutdown(srvlist* service) {
  static char const shutdown[]="/" SVCLEAN SVCLEAN_SHUTDOWN;
  int fd;

  if (mkdir(SVCLEAN, 0755)!=0 && errno!=EEXIST) {
    setstatus();
    printerr(err_create, service->name, "/" SVCLEAN);
    service->err=1;
    return;
  }

  fd=open(shutdown+1, O_WRONLY|O_CREAT, 0644);
  if (fd<0) {
    setstatus();
    printerr(err_create, service->name, shutdown);
    service->err=1;
    return;
  }

  close(fd);
}

#ifdef O_NDELAY
#define nonblock O_NDELAY
#else
#define nonblock O_NONBLOCK
#endif

static int svc_any_up,  svc_any_down;
static char svc_command[2];
static int svc_cmd_len;
static void svc(srvlist* service) {
  static char const control[]="/supervise/control";
  int fd;
  if (service->up==0) return;

  fd=open(control+1, O_WRONLY|nonblock);
  if (fd<0) {
    if (errno!=ENOENT && errno!=ENXIO) {
      setstatus();
      printerr(err_open, service->name, control);
      service->err=1;
    } else {
      service->up=0;
      svc_any_down=1;
    }
    return;
  }

  if (svc_cmd_len!=0) {
    buffer buf;
    char buf_space[sizeof svc_command];
    int const flags=fcntl(fd, F_GETFL, 0);

    if (flags==-1) {
      setstatus();
      printerr(err_getflags, null, null);
      service->err=1;
      return;
    }

    if (fcntl(fd, F_SETFL, flags|nonblock)==-1) {
      setstatus();
      printerr(err_setflags, null, null);
      service->err=1;
      return;
    }

    buffer_init(&buf, buffer_unixwrite, fd, buf_space, sizeof buf_space);
    if (buffer_putflush(&buf, svc_command, svc_cmd_len)!=0) {
      setstatus();
      printerr("write commands to ", service->name, control);
      close(fd);
      service->err=1;
      return;
    }
  }

  close(fd);
  svc_any_up=1;
}

typedef void iterfunc(srvlist*);
static iterfunc create_shutdown;
static iterfunc svc;

static void iterate(iterfunc* func) {
  srvlist* service;

  for (service=services; service!=null; service=service->next) {
    if (service->err) continue;

    if (fchdir(servicesdir)!=0)
      printerr_die(err_chdirto, servicesdirname, null);

    if (chdir(service->name)!=0) {
      if (errno!=ENOENT) {
        setstatus();
        printerr(err_chdirto, service->name, null);
      }
      service->islog=1;
      service->err=1;
      continue;
    }

    func(service);
  }
}

void do_svc(void) {
  iterate(svc);
  svc_cmd_len=0;
  for (;;) {
    do {
      svc_any_up=0;
      svc_any_down=0;
      iterate(svc);
      if (!svc_any_up) return;
    } while (svc_any_down);
    sleep(1);
  }
}

int main(int argc, char** argv) {
  static char const log[]="/log";
  DIR* dirh;
  srvlist* service, ** next;

  if (argc!=2) {
    buffer_puts(buffer_2, prog);
    buffer_puts(buffer_2, usage);
    buffer_puts(buffer_2, prog);
    buffer_puts(buffer_2, " /service\n");
    buffer_flush(buffer_2);
    _exit(100);
  }

  servicesdirname=argv[1];
  servicesdir=open(servicesdirname, O_RDONLY);
  if (servicesdir<0)
    printerr_die(err_open, servicesdirname, null);
  dirh=opendir(servicesdirname);
  if (dirh==null)
    printerr_die(err_open, servicesdirname, null);

  services=null;
  next=&services;
  for (;;) {
    struct dirent const* de;
    unsigned long namelen, loglen, size;

    errno=0;
    de=readdir(dirh);
    if (de==null) {
      if (errno==0) break;
      printerr_die("read ", servicesdirname, null);
    }

    if (de->d_name[0]=='.') continue;

    namelen=str_len(de->d_name);
    loglen=namelen+sizeof log;
    size=srvlist_size(loglen);
    if (loglen<sizeof log || size<loglen || size!=(unsigned int)size) {
      buffer_puts(buffer_2, prog);
      buffer_puts(buffer_2, err_);
      buffer_puts(buffer_2, "name too long: ");
      buffer_puts(buffer_2, de->d_name);
      buffer_puts(buffer_2, "\n");
      buffer_flush(buffer_2);
      _exit(100);
    }

    service=(srvlist*)alloc((unsigned int)size);
    if (service==null)
      printerr_die(err_alloc, null, null);

    service->next=null;
    service->up=1;
    service->err=0;
    service->islog=0;
    byte_copy(service->name, namelen, de->d_name);
    service->name[namelen]='\0';

    *next=service;
    next=&service->next;
  }
  closedir(dirh);

  iterate(create_shutdown);
  svc_command[0]='d';
  svc_command[1]='x';
  svc_cmd_len=2;
  do_svc();

  for (service=services; service!=null; service=service->next) {
    if (service->islog) continue;
    service->islog=1;
    service->up=1;
    service->err=0;
    byte_copy(service->name+str_len(service->name), sizeof log, log);
  }

  iterate(create_shutdown);
  svc_cmd_len=2;
  do_svc();
  sleep(6);
  svc_command[0]='x';
  svc_cmd_len=1;
  do_svc();

  _exit(status);
}
